@@ -8,7 +8,7 @@
 # License   : Revised BSD License
 # -----------------------------------------------------------------------------
 # Creation  : 26-Apr-2010
-# Last mod  : 06-Feb-2012
+# Last mod  : 13-Feb-2012
 # -----------------------------------------------------------------------------
 
 """
@@ -34,14 +34,14 @@ See also:
 - EC2, fabric, and "err: stdin: is not a tty"
   <http://blog.markfeeney.com/2009/12/ec2-fabric-and-err-stdin-is-not-tty.html>`_
 
-:copyright: (c) 2011 by Sébastien Pierre, see AUTHORS for more details.
+:copyright: (c) 2011,2012 by Sébastien Pierre.
 :license:   BSD, see LICENSE for more details.
 """
 
 import base64, bz2, crypt, hashlib, os, random, sys, re, string, tempfile, subprocess, types
 import fabric, fabric.api, fabric.operations, fabric.context_managers
 
-VERSION     = "0.2.1"
+VERSION     = "0.2.2"
 
 RE_SPACES   = re.compile("[\s\t]+")
 MAC_EOL     = "\n"
@@ -65,7 +65,7 @@ DEFAULT_OPTIONS = dict(
 def mode_local():
 	"""Sets Cuisine into local mode, where run/sudo won't go through
 	Fabric's API, but directly through a popen. This allows you to
-	easily test your Cuisine scripts without using Fabric"""
+	easily test your Cuisine scripts without using Fabric."""
 	global MODE_LOCAL
 	if MODE_LOCAL is False:
 		def custom_run( cmd ):
@@ -159,11 +159,11 @@ def run(*args, **kwargs):
 		return fabric.api.run(*args, **kwargs)
 
 def run_local(command):
-	"""A wrapper around subprocess"""
+	"""A wrapper around subprocess."""
 	pipe = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE).stdout
 	res = pipe.read()
 	# FIXME: Should stream the pipe, and only print it if fabric's properties allow it
-	print res
+	# print res
 	return pipe
 
 def sudo(*args, **kwargs):
@@ -178,28 +178,60 @@ def sudo(*args, **kwargs):
 #
 # =============================================================================
 
-def dispatch(function):
-	"""Dispatches the current function to specific implementation. For instance
-	@dispatch("package_ensure", select="package_system").
+def dispatch(prefix=None):
+	"""Dispatches the current function to specific implementation. The `prefix`
+	parameter indicates the common option prefix, and the `option_select()`
+	function will determine the function suffix.
+	
+	For instance the package functions are defined like that:
+
+	{{{
+	@dispatch("package")
+	def package_ensure(...):
+		...
+	def package_ensure_apt(...):
+		...
+	def package_ensure_yum(...):
+		...
+	}}}
+	
+	and then when a user does
+
+	{{{
+	cuisine.option_select("package", "yum")
+	cuisine.package_ensure(...)
+	}}}
+
+	then the `dispatch` function will dispatch `package_ensure` to 
+	`package_ensure_yum`.
+
+	If your prefix is the first word of the function name before the
+	first `_` then you can simply use `@dispatch` without parameters.
 	"""
-	def wrapper(*args, **kwargs):
-		function_name = function.__name__
-		prefix        = function_name.split("_")[0]
-		select        = fabric.api.env.get("option_" + prefix)
-		assert select, "No option defined for: %s, call select_%s(<YOUR OPTION>) to set it" % (prefix, prefix)
-		function_name = function.__name__ + "_" + select
-		specific      = eval(function_name)
-		if specific:
-			if type(specific) == types.FunctionType:
-				return specific(*args, **kwargs)
+	def dispatch_wrapper(function, prefix=prefix):
+		def wrapper(*args, **kwargs):
+			function_name = function.__name__
+			_prefix       = prefix or function_name.split("_")[0]
+			select        = fabric.api.env.get("option_" + _prefix)
+			assert select, "No option defined for: %s, call select_%s(<YOUR OPTION>) to set it" % (_prefix, prefix)
+			function_name = function.__name__ + "_" + select
+			specific      = eval(function_name)
+			if specific:
+				if type(specific) == types.FunctionType:
+					return specific(*args, **kwargs)
+				else:
+					raise Exception("Function expected for: " + function_name)
 			else:
-				raise Exception("Function expected for: " + function_name)
-		else:
-			raise Exception("Function variant not defined: " + function_name)
-	# We copy name and docstring
-	wrapper.__name__ = function.__name__
-	wrapper.__doc__  = function.__doc__
-	return wrapper
+				raise Exception("Function variant not defined: " + function_name)
+		# We copy name and docstring
+		wrapper.__name__ = function.__name__
+		wrapper.__doc__  = function.__doc__
+		return wrapper
+	if type(prefix) == types.FunctionType:
+		return dispatch_wrapper(prefix, None)
+	else:
+		return dispatch_wrapper
+
 
 # =============================================================================
 #
@@ -253,6 +285,8 @@ def text_ensure_line(text, *lines):
 	the end of it."""
 	eol = text_detect_eol(text)
 	res = list(text.split(eol))
+	if res[0] is '' and len(res) is 1:
+		res = list()
 	for line in lines:
 		assert line.find(eol) == -1, "No EOL allowed in lines parameter: " + repr(line)
 		found = False
@@ -346,12 +380,17 @@ def file_write(location, content, mode=None, owner=None, group=None):
 	os.write(fd, content)
 	# Upload the content if necessary
 	if not file_exists(location) or sig != file_sha256(location):
-		fabric.operations.put(local_path, location, use_sudo=(mode == MODE_SUDO))
+		if MODE == MODE_SUDO: mode = MODE_SUDO
+		try:
+			fabric.operations.put(local_path, location, use_sudo=(mode == MODE_SUDO))
+		except Exception, e:
+			print "cuisine.file_write exception:"
 	# Remove the local temp file
 	os.close(fd)
 	os.unlink(local_path)
-	# Ensure that the signature matches
+	# Ensures that the signature matches
 	assert sig == file_sha256(location)
+	file_attribs(location, mode=mode, owner=owner, group=group)
 
 def file_ensure(location, mode=None, owner=None, group=None, recursive=False):
 	"""Updates the mode/owner/group for the remote file at the given
@@ -413,7 +452,7 @@ def file_link(source, destination, symbolic=True, mode=None, owner=None, group=N
 	file_attribs(destination, mode, owner, group)
 
 def file_sha256(location):
-	"""Returns the SHA-256 sum (as a hex string) for the remote file at the given location"""
+	"""Returns the SHA-256 sum (as a hex string) for the remote file at the given location."""
 	return run('sha256sum "%s" | cut -d" " -f1' % (location))
 
 # TODO: From McCoy's version, consider merging
@@ -437,18 +476,13 @@ def dir_exists(location):
 
 def dir_ensure(location, recursive=False, mode=None, owner=None, group=None):
 	"""Ensures that there is a remote directory at the given location,
-	optionnaly updating its mode/owner/group.
+	optionally updating its mode/owner/group.
 
 	If we are not updating the owner/group then this can be done as a single
 	ssh call, so use that method, otherwise set owner/group after creation."""
-	if mode:
-		mode_arg = "-m %s" % (mode)
-	else:
-		mode_arg = ""
-	run('test -d "%s" || mkdir %s %s "%s" && echo OK ; true' %
-		(location, recursive and "-p" or "", mode_arg, location))
-	if owner or group:
-		dir_attribs(location, owner=owner, group=group)
+	run('test -d "%s" || mkdir %s "%s" && echo OK ; true' % (location, recursive and "-p" or "", location))
+	if owner or group or mode:
+		dir_attribs(location, owner=owner, group=group, mode=mode)
 
 # =============================================================================
 #
@@ -456,6 +490,10 @@ def dir_ensure(location, recursive=False, mode=None, owner=None, group=None):
 #
 # =============================================================================
 
+@dispatch
+def package_upgrade():
+	"""Updates every package present on the system."""
+
 @dispatch
 def package_update(package=None):
 	"""Updates the package database (when no argument) or update the package
@@ -463,18 +501,25 @@ def package_update(package=None):
 
 @dispatch
 def package_install(package, update=False):
-	"""Installs the given package/list of package, optionnaly updating
+	"""Installs the given package/list of package, optionally updating
 	the package database."""
 
 @dispatch
-def package_ensure(package):
-	"""Tests if the given package is installed, and installes it in
-	case it's not already there."""
+def package_ensure(package, update=False):
+	"""Tests if the given package is installed, and installs it in
+	case it's not already there. If `update` is true, then the
+	package will be updated if it already exists."""
 
 # -----------------------------------------------------------------------------
 # APT PACKAGE (DEBIAN/UBUNTU)
 # -----------------------------------------------------------------------------
 
+def repository_ensure_apt(repository):
+	sudo("add-apt-repository " + repository)
+
+def package_upgrade_apt():
+	sudo("apt-get --yes upgrade")
+
 def package_update_apt(package=None):
 	if package == None:
 		sudo("apt-get --yes update")
@@ -491,12 +536,13 @@ def package_install_apt(package, update=False):
 	sudo("apt-get --yes install %s" % (package))
 
 
-def package_ensure_apt(package):
+def package_ensure_apt(package, update=False):
 	status = run("dpkg-query -W -f='${Status}' %s ; true" % package)
 	if status.find("not-installed") != -1 or status.find("installed") == -1:
 		package_install(package)
 		return False
 	else:
+		if update: package_update(package)
 		return True
 
 # =============================================================================
@@ -532,16 +578,13 @@ def user_create(name, passwd=None, home=None, uid=None, gid=None, shell=None,
 	"""Creates the user with the given name, optionally giving a
 	specific password/home/uid/gid/shell."""
 	options = ["-m"]
-	if passwd:
-		method = 6
-		saltchars = string.ascii_letters + string.digits + './'
-		salt = ''.join([random.choice(saltchars) for x in range(8)])
-		passwd_crypted = crypt.crypt(passwd, '$%s$%s' % (method, salt))
-		options.append("-p '%s'" % (passwd_crypted))
 	if home:
 		options.append("-d '%s'" % (home))
 	if uid:
 		options.append("-u '%s'" % (uid))
+	#if group exists already but is not specified, useradd fails
+	if not gid and group_check(name):
+		gid = name
 	if gid:
 		options.append("-g '%s'" % (gid))
 	if shell:
@@ -551,6 +594,9 @@ def user_create(name, passwd=None, home=None, uid=None, gid=None, shell=None,
 	if uid_max:
 		options.append("-K UID_MAX='%s'" % (uid_max))
 	sudo("useradd %s '%s'" % (" ".join(options), name))
+	if passwd:
+		sudo("echo %s:%s | chpasswd" % (name, passwd))
+
 
 def user_check(name):
 	"""Checks if there is a user defined with the given name,
@@ -562,6 +608,7 @@ def user_check(name):
 	results = {}
 	if d:
 		d = d.split(":")
+		assert len(d) >= 7, "/etc/passwd entry is expected to have at least 7 fields, got %s in: %s" % (len(d), d.join(":"))
 		results = dict(name=d[0], uid=d[2], gid=d[3], home=d[5], shell=d[6])
 	if s:
 		results['passwd'] = s
@@ -578,18 +625,6 @@ def user_ensure(name, passwd=None, home=None, uid=None, gid=None, shell=None):
 		user_create(name, passwd, home, uid, gid, shell)
 	else:
 		options = []
-		if passwd != None and d.get('passwd') != None:
-			method, salt = d.get('passwd').split('$')[1:3]
-			passwd_crypted = crypt.crypt(passwd, '$%s$%s' % (method, salt))
-			if passwd_crypted != d.get('passwd'):
-				options.append("-p '%s'" % (passwd_crypted))
-		if passwd != None and d.get('passwd') is None:
-			# user doesn't have passwd
-			method = 6
-			saltchars = string.ascii_letters + string.digits + './'
-			salt = ''.join([random.choice(saltchars) for x in range(8)])
-			passwd_crypted = crypt.crypt(passwd, '$%s$%s' % (method, salt))
-			options.append("-p '%s'" % (passwd_crypted))
 		if home != None and d.get("home") != home:
 			options.append("-d '%s'" % (home))
 		if uid != None and d.get("uid") != uid:
@@ -600,6 +635,9 @@ def user_ensure(name, passwd=None, home=None, uid=None, gid=None, shell=None):
 			options.append("-s '%s'" % (shell))
 		if options:
 			sudo("usermod %s '%s'" % (" ".join(options), name))
+		if passwd:
+			encoded_password = base64.b64encode("%s:%s" % (name, passwd))
+			sudo("echo %s | base64 --decode | chpasswd" % (encoded_password,name, passwd))
 
 # =============================================================================
 #
@@ -697,13 +735,14 @@ def ssh_authorize(user, key):
 		else:
 			return True
 	else:
-		file_write(keyf, key)
+		# Make sure that .ssh directory exists, see #42
+		dir_ensure(os.path.dirname(keyf), user=user, group=user, mode="700")
+		file_write(keyf, key,             user=user, group=user, mode="600")
 		return False
 
-
 def upstart_ensure(name):
 	"""Ensures that the given upstart service is running, restarting
-	it if necessary"""
+	it if necessary."""
 	status = sudo("service %s status" % name)
 	if status.find("is running") >= 0 or status.find("/running") >= 0:
 		sudo("service %s restart" % name)
@@ -724,8 +763,24 @@ def system_uuid():
 	"""Gets a machines UUID (Universally Unique Identifier)."""
 	return sudo('dmidecode -s system-uuid | tr "[A-Z]" "[a-z]"')
 
+#Only tested on Ubuntu!
+def locale_check(locale):
+	locale_data = sudo("locale -a | egrep '^%s$' ; true" % (locale,))
+	return locale_data == locale
+
+def locale_ensure(locale):
+	if locale_check(locale):
+		return
+	with fabric.context_managers.settings(warn_only=True):
+		sudo("/usr/share/locales/install-language-pack %s" % (locale,))
+	sudo("dpkg-reconfigure locales")
+
+
 # Sets up the default options so that @dispatch'ed functions work
 for option, value in DEFAULT_OPTIONS.items():
 	eval("select_" + option)(value)
 
+
+
+
 # EOF - vim: ts=4 sw=4 noet
